"""ZWiki product

This product allows you to build wiki webs with zope objects.
It includes a sample wikiweb to help you get started quickly.
There is a demo & documentation site running at:

 http://joyful.com/zwiki/

You are responsible for security; zwiki may contain unforeseen risks.
I appreciate all feedback, bugreports, fixes, responses to the ??
comments in the code, re-implementations, etc.

(c) 1999-2000 Simon Michael <simon@joyful.com> & contributors
This product is available under the GPL.
All rights reserved, all disclaimers apply, etc.

?? marks questions and dubious areas
"""

__version__="$Revision: 0.7 $"[11:-2]

from types            import *
from string           import split,join, find, lower
from urllib           import quote, unquote
import re

import Acquisition
from DateTime         import DateTime
from Globals          import HTMLFile, MessageDialog
from DocumentTemplate import HTML
from OFS.DTMLDocument import DTMLDocument
from OFS.Document     import Document
from App.Common       import absattr
from DocumentTemplate.DT_Util import html_quote
from StructuredText   import html_with_references

intl_char_entities = (
    ('\300', '&Agrave;'),               ##
    ('\302', '&Acirc;'),                ##
    ('\311', '&Eacute;'),               ##
    ('\312', '&Ecirc;'),                ##
    ('\316', '&Icirc;'),                ##
    ('\324', '&Ocirc;'),                ##
    ('\333', '&Ucirc;'),                ##
    ('\340', '&agrave;'),               ##
    ('\342', '&acirc;'),                ##
    ('\347', '&ccedil;'),               ##
    ('\350', '&egrave;'),               ##
    ('\351', '&eacute;'),               ##
    ('\352', '&ecirc;'),                ##
    ('\356', '&icirc;'),                ##
    ('\364', '&ocirc;'),                ##
    ('\371', '&ugrave;'),               ##
    ('\373', '&ucirc;'),                ##
)

# Regular Expressions

strip_cr         = r'\r(\n)'
bracketedexpr    = r'\[([^\]0-9][^]]*)\]' # trying to co-exist with stx's references
underlinedexpr   = r'_([^_]+)_'
urllink          = r'["=]?((http|ftp|mailto|file):[A-Za-z0-9/:@_%~#=\.\-\?]+)'
wikiname1        = r'\b[A-Z]+[a-z]+[A-Z][A-Za-z]*[0-9]*'
wikiname2        = r'\b[A-Z][A-Z]+[a-z][A-Za-z]*[0-9]*'
simplewikilink   = r'!?(%s|%s)' % (wikiname1, wikiname2)
wikilink         = r'!?(%s|%s|%s|%s)' % (wikiname1,
                                         wikiname2,
                                         bracketedexpr,
                                         urllink)
interwikilink    = r'!?((?P<local>%s):(?P<remote>[\w]+))' % (wikilink)
remotewikiurl    = r'(?m)RemoteWikiURL[:\s]*(.*)$'
protected_line   = r'(?m)^!(.*)$'
hrule            = re.compile(r'^----[-]*', re.MULTILINE);

DEFAULT_PAGE_TYPE = 'structuredtext'

class ZWikiPage(DTMLDocument):
    """
	A ZWikiPage is a DTML Document which knows how to render itself in various 
    wiki styles, and provides some utility methods to support wiki-building.
    """
    meta_type = "ZWiki Page"
    icon      = "misc_/ZWiki/ZWikiPage_icon"
    page_type = DEFAULT_PAGE_TYPE
    parents   = None
    assignedTo = ''
    status     = ''
 
    # how would you append to the superclass' _properties here ??
    _properties=({'id':'title',      'type':'string', 'mode':'w'},
                 {'id':'page_type',  'type':'string', 'mode':'w'},
                 {'id':'parents',    'type':'lines',  'mode':'w'},
                 {'id':'assignedTo', 'type':'string', 'mode':'w'},
                 {'id':'status',     'type':'string', 'mode':'w'}
                )

    __ac_permissions__=(
        ('Change ZWiki Pages',  ('edit','PUT','manage_edit','manage_upload')), 
        ('Approve ZWiki Pages', ()),
        ('Manage ZWiki Pages',  ()),
        )
           
    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render a zwiki page:  there are four possible renderings:
           r=frm|null:  generate the frameset
           r=top:       generate the top frame
           r=left:      generate the left frame
           r=body:      generate the body
        """
        # define this here so header & footer methods can get at the page attributes.
        kw['this']=self

        # backwards compatibility with older representation of parents
        if (not hasattr(self, 'parents')) or self.parents is None:
            self.parents = []

        if not hasattr(self, 'page_type'):
            self.page_type = DEFAULT_PAGE_TYPE
        elif self.page_type == 'structuredtext_dtml':
            self.page_type = 'structuredtextdtml'
        elif self.page_type == 'Structured Text':
            self.page_type = 'structuredtext'
        elif lower(self.page_type[0:4]) == 'html':
            self.page_type = 'html'
        elif self.page_type == 'Plain Text':
            self.page_type = 'plaintext'
        elif self.page_type != 'structuredtextdtml' and self.page_type != 'structuredtext' and self.page_type != 'html' and self.page_type != 'plaintext':
            self.page_type = DEFAULT_PAGE_TYPE

        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/html')
       
        # get either the raw text or the dtml-processed text
        if find(self.page_type, 'dtml') > -1:
            text = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        else:
            text = self.raw

        # divide the text contents into the top, left, and main sections
        text = split(text, "~~")
         
        if REQUEST is None or not REQUEST.form.has_key('r') or REQUEST.form['r']=='frm':
            return apply(self.standard_frameset.__call__,
                         (client, REQUEST, RESPONSE), kw)
        elif REQUEST.form['r']=='top':
            return self.generateSection(text, -3, 'standard_top_header', 'standard_top_footer', 
                                        kw, client, REQUEST, RESPONSE)            
        elif REQUEST.form['r']=='left':
            return self.generateSection(text, -2, 'standard_left_header', 'standard_left_footer', 
                                        kw, client, REQUEST, RESPONSE)            
        else:
            return self.generateSection(text, -1, 'standard_wiki_header', 'standard_wiki_footer', 
                                        kw, client, REQUEST, RESPONSE)

    def generateSection(self, text, section_index, header_name, footer_name, kw, client=None, REQUEST=None, RESPONSE=None):
        # prepend header
        if REQUEST is not None:
            if hasattr(self, header_name):
                hdr = apply(getattr(self, header_name).__call__,
                            (client, REQUEST, RESPONSE), kw)
            else:
                hdr = default_wiki_header
        else: hdr = ''

        # generate the body
        try:
            if self.page_type == 'structuredtextdtml' or self.page_type == 'structuredtext':
                body = text[section_index]
                body = hrule.sub('<hr>', body)
                body = str(html_with_references(body, level=2))
                body = re.sub(protected_line, self._protect_line, body)
                body = re.sub(interwikilink, self._interwikilink_replace, body)
                body = re.sub(wikilink, self._wikilink_replace, body)
            elif self.page_type == 'html':
                body = text[section_index]
                body = re.sub(protected_line, self._protect_line, body)
                body = re.sub(interwikilink, self._interwikilink_replace, body)
                body = re.sub(wikilink, self._wikilink_replace, body)
            elif self.page_type == 'plaintext':
                body = "<pre>\n" + html_quote(text[section_index]) + "\n</pre>\n"
        except:
            body = ''
        
        # append footer
        if REQUEST is not None:
            if hasattr(self, footer_name):
                ftr = apply(getattr(self, footer_name).__call__,
                            (client, REQUEST, RESPONSE), kw)
            else:
                ftr = default_wiki_footer
        else: ftr = ''

        return hdr + body + ftr

    def _protect_line(self, match):
        """
        Protect an entire line from _wikilink_replace, by protecting all 
        its wikilink occurrences
        """
        return re.sub(wikilink, r'!\1', match.group(1))
    
    def _wikilink_replace(self, match):
        """replace an occurrence of wikilink with a suitable hyperlink"""
        # matches beginning with ! should be left alone
        if re.match('^!',match.group(0)):
            return match.group(1)

        m = match.group(1)

        # strip any enclosing [] or __
        # display these as a visual cue ??
        if re.match(bracketedexpr,m):
            m = re.sub(bracketedexpr, r'\1', m)
        elif re.match(underlinedexpr,m):
            m = re.sub(underlinedexpr, r'\1', m)
        
        if re.match(urllink,m):
            if re.match('^["=]',m):
                return m
            else:
                return '<a target="_top" href="%s">%s</a>' % (m, m)
        elif hasattr(self.aq_parent, m):
            # if something of this name exists, link to it;
            return '<a target="_top" href="%s">%s</a>' % (quote(m), m)
        else:
            # otherwise, provide a "?" creation link
            return '%s<a target="_top" href="%s/new?id=%s">?</a>' % (m, quote(self.id()), quote(m))
        
    def _interwikilink_replace(self, match):
        """replace an occurrence of interwikilink with a suitable hyperlink"""
        # matches beginning with ! should be left alone
        if re.match('^!', match.group(0)): return match.group(1)

        localname  = match.group('local')
        remotename = match.group('remote') # named groups come in handy here!

        # look for a RemoteWikiURL definition
        if hasattr(self.aq_parent, localname): 
            localpage = getattr(self.aq_parent,localname)
            m = re.search(remotewikiurl, str(localpage))
            if m is not None:
                remoteurl = m.group(1)
                
                # we have a valid inter-wiki link
                link = '<a target="_top" href="%s%s">%s:%s</a>' % \
                       (remoteurl, remotename, localname, remotename)
                # protect it from any further wiki-ization
                return re.sub(wikilink, r'!\1', link)

        # otherwise, leave alone
        return match.group(0)

    def new(self, id, title='', REQUEST=None, submit=None):
        """Create a new ZWikiPage and display it."""
        self._validateProxy(REQUEST)

        # set up default content
        if hasattr(self.aq_parent, 'standard_wiki_page'):
            t = apply(self.aq_parent.standard_wiki_page.__call__,
                      (None, REQUEST, REQUEST.RESPONSE))
        else:
            t = default_wiki_page

        # create page
        ob = ZWikiPage(t, __name__=id)

        ob.title = title
        id = self.aq_parent._setObject(id, ob)

        if ob.parents is None:
            ob.parents = [self.id()]
        else:
            ob.parents.append(self.id())

        ob.page_type  = DEFAULT_PAGE_TYPE
        ob.assignedTo = ''
        ob.status     = ''

        if REQUEST is not None:
            # redirect browser to the new page
            # all these redirects.. better/safer way ??
            u=REQUEST['URL2']
            REQUEST.RESPONSE.redirect(u + '/' + quote(id))

    def manage_FTPget(self):
        """Get source for FTP download"""
        return "Wiki-Safetybelt: %s\n\n%s" % (
            self.editTimestamp(), self.read())

    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        body=REQUEST.get('BODY', '')
        self._validateProxy(REQUEST)

        m = re.match(r'Wiki-Safetybelt: ([0-9]+[.][0-9]+)[\n][\n]?', body)
        if m:
            self.checkEditTimeStamp(m.group(1))
            body=body[m.span()[1]:]            

        self.munge(body)
        RESPONSE.setStatus(204)
        return RESPONSE

    def checkEditTimeStamp(self, timeStamp):
        if timeStamp is not None and timeStamp != self.editTimestamp():
            raise 'EditingConflict', (
                '''Someone has edited this page since you loaded the
                page for editing.<p>
                Try editing the page again.
                ''')
                                                            
    def editTimestamp(self): 
        return str(self._p_mtime)

    def edit(self, text=None, page_type=None, assignedTo=None, status=None, REQUEST=None, submit=None, timeStamp=None):
        """change a ZWikiPage's content, page_type, status, etc."""
        self._validateProxy(REQUEST)

        self.checkEditTimeStamp(timeStamp)

        if page_type is not None:
            self.page_type = page_type

        if assignedTo is not None:
            self.assignedTo = assignedTo

        if status is not None:
            self.status = status

        if text is not None:
            text = self._edit_cleanups(text)
            if find(self.page_type, 'dtml') > -1:
                # for DTML page types, test for DTML problems - zope
                # will roll back this transaction, and the user will
                # get an appropriate error
                self.munge(text)
                apply(DTMLDocument.__call__,(self, self.aq_parent,
                                             REQUEST, REQUEST.RESPONSE))
            else:
                self.raw = text
        
        if REQUEST is not None:
            # force a browser page refresh
            try: u=self.DestinationURL()
            except: u=REQUEST['URL1']
            REQUEST.RESPONSE.redirect(u)

    def _edit_cleanups(self, t=None):
        """do some cleanups on new page text"""
        # strip any browser-appended ^M's
        t = re.sub(strip_cr, r'\1', t)

        # convert international characters to HTML entities for safekeeping
        for c,e in intl_char_entities:
            t = re.sub(c, e, t)

        return t
        
    src__roles__ = None

    def getText(self, REQUEST=None, RESPONSE=None):
        """Permission-free version of zwiki page text"""
        return self.read()

    def isPageType(self, page_type, REQUEST=None, RESPONSE=None):
        if hasattr(self, 'page_type'):
            if self.page_type == page_type:
                return 'CHECKED'
        else:
            if page_type == 'structuredtext':
                return 'CHECKED'
        return ''

    def getStatus(self, REQUEST=None, RESPONSE=None):
        """Permission-free version of zwiki page status"""
        if hasattr(self, 'status'):
            return self.status
        else:
            return ''

    def getAssignedTo(self, REQUEST=None, RESPONSE=None):
        """Permission-free version of zwiki page assignment"""
        if hasattr(self, 'assignedTo'):
            return self.assignedTo
        else:
            return ''

    # url utilities to make links easier
    def wiki_page_url(self):
        """return the url path for the current wiki page"""

        o = self
        url = []
        while hasattr(o,'id'):
            url.insert(0,absattr(o.id))
            o = getattr(o, 'aq_parent', None)
        return quote('/' + join(url[1:],'/'))

    def wiki_base_url(self):
        """return the base url path for the current wiki"""

        u = self.wiki_page_url()
        u = split(u,'/')
        u = u[:-1]
        u = join(u,'/')
        return quote(u)

    # only access unprotected resources when using DTML
    def validate(self, inst, parent, name, value, md):
         try:
             if (name[:3]=='aq_' and
                 name != 'aq_parent' and name != 'aq_explicit'):
                 return 0
         except: pass # name might not be a string!
 
         # Try to get roles
         if hasattr(value, '__roles__'): roles=value.__roles__
         else:
             if hasattr(parent,'__roles__'): roles=parent.__roles__
             elif hasattr(parent, 'aq_acquire'):
                 try: roles=parent.aq_acquire('__roles__')
                 except AttributeError:
                     if hasattr(inst, 'aq_base'): inst=inst.aq_base
                     if hasattr(parent, 'aq_base'): parent=parent.aq_base
                     return inst is parent
             else:
                 if hasattr(inst, 'aq_base'): inst=inst.aq_base
                 if hasattr(parent, 'aq_base'): parent=parent.aq_base
                 return inst is parent
             value=parent
             
         if roles is None or ("Anonymous" in roles): return 1
 
         if inst is parent:
             raise 'Unauthorized', ('You are not authorized to access <em>%s</em>.' % name)
 
         return 0
 
    def reparent(self, parents=None, REQUEST=None):
        """Reset parents property according to request."""
        if parents is None:
            parents = REQUEST.get('parents', None)
        if parents is None:
            self.parents = []
        else:
            if type(parents) != ListType:
                parents = [parents]
            self.parents = parents
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(REQUEST['URL1'] + "/backlinks")

    def updateparentage(self, oldname, newname=None):
        """
        Replace oldname with newname in all parent information.
        This method is often used after deleting a page, use newname=None
        """
        container = self.aq_parent
        pages = container.objectValues(spec='ZWiki Page')
        for p in pages:
            parents = p.parents
            changed = 0
            if type(parents) == ListType:
                if parents.count(oldname) > 0:
                    parents.remove(oldname)
                    if newname is not None:
                        parents.append(newname)
                    changed = 1
            elif parents == oldname:
                if newname is not None:
                    parents = newname
                else:
                    parents = []
                changed = 1
            if changed == 1:
                p.parents = parents

    def get_ancestors(self):
        """Return a trimmed nesting structure indicating this page's ancestors."""
        container = self.aq_parent
        ancestors = {}
        offspring = {}
        tops = {}                       # Ancestors that have no parents.
        todo = {self.id(): None}
        while todo:
            doing = todo
            todo = {}
            for i in doing.keys():
                if ancestors.has_key(i):
                    continue            # We've already collected this one.
                else:
                    if not hasattr(container, i):
                        continue        # Absent - ignore it.
                    ancestors[i] = None
                    obj = container[i]
                    if not hasattr(obj, 'parents'):
                        continue
                    parents = obj.parents
                    if type(parents) != ListType:
                        parents = []
                    if parents:  # non-empty test
                        for p in parents:
                            if offspring.has_key(p):
                                offspring[p].append(i)
                            else:
                                offspring[p] = [i]
                            todo[p] = None
                    else: 
                        tops[i] = None

        # Ok, now go back down, unravelling each forebear only once:
        tops = tops.keys()
        tops.sort
        did = {}; got = []
        for t in tops:
            got.append(descend_ancestors(t, ancestors, did, offspring))
        return got

    def ancestors(self, REQUEST=None, width=150, styleclass=None):
        result = self.get_ancestors()
        return present_nestingboxes(None, result, self.wiki_base_url(), width, styleclass)

    def context(self, REQUEST=None):
        """Return HTML showing this page's parents and siblings."""
        myid = self.id()
        result = WikiNesting(self.aq_parent).get_up_and_back(myid)
        return present_nesting(myid, result, self.wiki_base_url())
                
    def offspring(self, REQUEST=None, RESPONSE=None):
        """Return a presentation of all my offspring."""
        myid = self.id()
        result = WikiNesting(self.aq_parent).get_offspring([myid])
        return present_nesting(myid, result, self.wiki_base_url())

    def map(self, REQUEST=None, RESPONSE=None):
        """
        Present the nesting layout of the entire wiki, showing:

        - All the independent nodes, ie those without parents or children, and

        - All the branches in the wiki - from the possibly multiple roots.
        """
        map = WikiNesting(self.aq_parent).get_map()
        singletons = []
        combos = []
        rel = self.wiki_base_url()
        for i in map:
            if type(i) == StringType:
                singletons.append('<a href="%s/%s">%s</a>'
                                  % (rel, quote(i), i))
            else:
                combos.append(i)
        
        return "<h3>Singletons</h3> <ul>%s</ul><h3>The rest</h3>%s</p>" % (
                join(singletons, " "), present_nesting(self, combos, rel))
        
    def __repr__(self):
        return ("<%s %s at 0x%s>"
                % (self.__class__.__name__, `self.id()`, hex(id(self))[2:]))

class WikiNesting(Acquisition.Implicit):
    """
    Given a wiki directory, generate nesting relationship from parents info.

    In a nesting, nodes are represented as:
       - Leaves: the string name of the page
       - Nodes with children: a list beginning with the parent node's name
       - Nodes with omitted children (for brevity): list with one string.
    """
    # XXX We could make this a persistent object and minimize recomputes.
    #     Put it in a standard place in the wiki's folder, or have the
    #     wikis in a folder share an instance, but use a single
    #     persistent one which need not recompute all the relationship
    #     maps every time - just needs to compare all pages parents
    #     settings with the last noticed parents settings, and adjust
    #     the children, roots, and parents maps just for those that
    #     changed.  On this first cut we just recompute it all...

    def __init__(self, container):
        self.container = container
        self.set_nesting()

    def set_nesting(self):
        """Preprocess for easy derivation of nesting structure.

        We set:
          - .parentmap: {'node1': ['parent1', ...], ...}
          - .childmap:  {'node1': ['child1', ...], ...}
          - .roots: {'root1': None, ...}"""

        pages = self.container.objectValues(spec='ZWiki Page')
        pagenames = []
        parentmap = {}                  # 'node': ['parent1', ...]
        childmap = {}                   # 'node': ['child1', ...]
        roots = {}
        for pg in pages:
            parents = pg.parents
            if type(parents) != ListType: # SKWM I think ken's zope remembers [] better
                parents = []
            pgnm = pg.id()
            pagenames.append(pgnm)

            # Parents is direct:
            parentmap[pgnm] = parents[:]

            # We have a root if node acknowledges no parents:
            if not parents:
                roots[pgnm] = None

            if not childmap.has_key(pgnm):
                childmap[pgnm] = []

            # Register page as child for all its parents:
            for p in parents:
                if childmap.has_key(p): pchildren = childmap[p]
                else: pchildren = []
                if pgnm not in pchildren: pchildren.append(pgnm)
                childmap[p] = pchildren

        for k in parentmap.keys():
            parentmap[k].sort()
        for k in childmap.keys():
            childmap[k].sort()
        self.parentmap = parentmap
        self.childmap = childmap
        self.roots = roots

    def get_map(self):
        """Return nesting of entire wiki."""
        did = {}
        got = []
        roots = self.roots.keys()
        roots.sort()
        return self.get_offspring(roots)

    def get_offspring(self, pages, did=None):
        """
		Return nesting showing all offspring of a list of wiki pagenames.
        did is used for recursion, to prune already elaborated pages.
		"""
        if did is None: did = {}
        got = []
        for p in pages:
            been_there = did.has_key(p)
            did[p] = None
            if self.childmap.has_key(p):
                children = self.childmap[p]
                if children:
                    subgot = [p]
                    if not been_there:
                        subgot.extend(self.get_offspring(children, did=did))
                    got.append(subgot)
                else:
                    got.append(p)
            else:
                got.append(p)
        return got

    def get_up_and_back(self, pagename):
        """
		Return nesting showing page containment and immediate offspring.
		"""
        # Go up, identifying all and topmost forbears:
        ancestors = {}                  # Ancestors of pagename
        tops = {}                       # Ancestors that have no parents
        todo = {pagename: None}
        parentmap = self.parentmap
        while todo:
            doing = todo
            todo = {}
            for i in doing.keys():
                if ancestors.has_key(i):
                    continue            # We already took care of this one.
                else:
                    ancestors[i] = None
                    if parentmap.has_key(i):
                        parents = parentmap[i]
                    if parents:
                        for p in parents:
                            todo[p] = None
                    else: tops[i] = None

        ancestors[pagename] = None      # Finesse inclusion of page's offspring

        # Ok, now go back down, showing offspring of all intervening ancestors:
        tops = tops.keys()
        tops.sort
        did = {}; got = []
        childmap = self.childmap
        for t in tops:
            got.append(descend_ancestors(t, ancestors, did, childmap))
        return got

def present_nesting(myid, nesting, rel):
    """Present an HTML outline of a nesting structure.

    See WikiNesting docstring for nesting structure description."""

    result = []
    result.append('<ul>')
    _present_nesting_aux(myid, nesting, rel, result)
    result.append('</ul>')

    return join(result, "\n")
        
def _present_nesting_aux(myid, nesting, rel, result):
    for n in nesting:
        if type(n) == ListType:
            result.append(_present_listitem(n[0], myid, rel))
            
            if len(n) > 1:
                result.append('<ul>')
                for i in n[1:]:
                    if type(i) == ListType:
                        _present_nesting_aux(myid, [i], rel, result)
                    else:
                        result.append(_present_listitem(i, myid, rel))
                result.append('</ul>')
        else:
            result.append(_present_listitem(n, myid, rel))
    return result

def _present_listitem(n, myid, rel):
    if n == myid:
        # Distinguish that entry is for specified node.
        nodenm = '<strong>%s</strong>' % n
    else: nodenm = n

    return '<li><a target="_top" href="%s/%s">%s</a>' % (rel, quote(n), nodenm)

def present_nestingboxes(myid, nesting, rel, width=150, styleclass=None):
    """Present a vertically stacked boxes view of a nesting structure.

    See WikiNesting docstring for nesting structure description."""

    result = []

    result.append('<table width="%s" bgcolor=black cellspacing=1>' % (width))
    _present_nestingboxes_aux(myid, nesting, None, rel, '', result, styleclass)
    result.append('</table>')

    return join(result, "\n")

def _present_nestingboxes_aux(myid, nesting, parentname, rel, indent, result, styleclass):
    for n in nesting:
        if type(n) == ListType:
            result.append(_present_box(n[0], parentname, indent, myid, rel, styleclass))
            
            if len(n) > 1:
                for i in n[1:]:
                    if type(i) == ListType:
                        _present_nestingboxes_aux(myid, [i], n[0], rel, indent+'&nbsp;&nbsp;', result, styleclass)
                    else:
                        result.append(_present_box(i, n[0], indent+'&nbsp;&nbsp;', myid, rel, styleclass))
        else:
            result.append(_present_box(n, parentname, indent, myid, rel, styleclass))
    return result

def _present_box(n, parentname, indent, myid, rel, styleclass):
    name = n[:] # make copy

    # remove prefix if same as parent name
    if parentname is not None:
        if find(name, parentname) == 0:
            name = '...' + name[len(parentname):]

    # indicate that entry is for specified node
    if n == myid:
        name = '<strong>%s</strong>' % name

    if styleclass is not None:
        #return '<tr><td><div class="%s">%s<a target="_top" href="%s/%s">%s</a></div></td></tr>' % (styleclass, indent, rel, quote(n), name)
        return '<tr><td class="%s">%s<a target="_top" href="%s/%s">%s</a></td></tr>' % (styleclass, indent, rel, quote(n), name)
    else:
        return '<tr><td>%s<a target="_top" href="%s/%s">%s</a></td></tr>' % (indent, rel, quote(n), name)

def descend_ancestors(page, ancestors, did, children):
    """Create nesting of ancestors leading to page.

    page is the name of the subject page.
    ancestors is a mapping whose keys are pages that are ancestors of page
    children is a mapping whose keys are pages children, and the values
       are the children's parents.

    Do not repeat ones we already did.
    """
    got = []
    for c in ((children.has_key(page) and children[page]) or []):
        if not ancestors.has_key(c):
            # We don't descend offspring that are not ancestors.
            got.append(c)
        elif ((children.has_key(c) and children[c]) or []):
            if did.has_key(c):
                # We only show offspring of ancestors once...
                got.append([c])
            else:
                # ... and this is the first time.
                did[c] = None
                got.append(descend_ancestors(c, ancestors, did, children))
        else:
            got.append(c)
    got.sort()                  # Terminals will come before composites.
    got.insert(0, page)
    return got

def manage_addZWikiPage(self, id, title='', file='', REQUEST=None, submit=None):
    """
    Add a ZWiki Page object with the contents of file. 
    If 'file' is empty, default text is used.
    """
    if type(file) is not StringType: file=file.read()

    if not file:
        if hasattr(self, 'standard_wiki_page'):
            file = absattr(self.standard_wiki_page)
        else:
            file = default_wiki_page

    ob=ZWikiPage(file, __name__=id)
    ob.title=title
    id=self._setObject(id, ob)

    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''

manage_addZWikiPageForm=HTMLFile('zwikipageAdd', globals())

default_wiki_page=\
'''This is a new ZWiki page.
'''

default_wiki_header=HTML(source_string=\
'''<HTML>
<HEAD>
<TITLE>zwiki: <dtml-var id></TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF">
<table width="100%" border="0" cellspacing="0" cellpadding="3"  bgcolor="eeeeee"><tr>
<td valign="middle">
<h1><dtml-var id></h1>
</td>
</tr>
</table>
''')

default_wiki_footer=HTML(source_string=\
'''
</BODY>
</HTML>
''')

default_minimal_header=HTML(source_string=\
'''<HTML><BODY>
''')

default_minimal_footer=HTML(source_string=\
'''</BODY></HTML>
''')